#!/usr/bin/env python3
# coding: utf-8

import base64
import hashlib
import mimetypes
import os.path
import tarfile
import urllib.parse
import zipfile
from functools import cached_property
from typing import Generator, List, Union, Iterable


def read_as_chunks(path: str, length=-1, offset=0, chunksize=65536) \
        -> Generator[bytes, None, None]:
    if length == 0:
        return
    if length < 0:
        length = float('inf')
    chunksize = min(chunksize, length)
    with open(path, 'rb') as fin:
        fin.seek(offset)
        while chunksize:
            chunk = fin.read(chunksize)
            if not chunk:
                break
            yield chunk
            length -= chunksize
            chunksize = min(chunksize, length)


def compute_checksum(path_or_chunks: Union[str, Iterable[bytes]], algo='sha1'):
    hashobj = hashlib.new(algo) if isinstance(algo, str) else algo
    # path_or_chunks:str - a path
    if isinstance(path_or_chunks, str):
        chunks = read_as_chunks(path_or_chunks)
    else:
        chunks = path_or_chunks
    for chunk in chunks:
        hashobj.update(chunk)
    return hashobj


def checksum(path: str, algo='sha1', length=-1, offset=0):
    chunks = read_as_chunks(path, length=length, offset=offset)
    return compute_checksum(chunks, algo=algo)


def checksum_hexdigest(path: str, algo='sha1', length=-1, offset=0):
    hashobj = checksum(path, algo=algo, length=-1, offset=0)
    return hashobj.hexdigest()


def b64_encode_data_url(mediatype: str, content: bytes):
    b64 = base64.b64encode(content).decode('ascii')
    return 'data:{};base64,{}'.format(mediatype, b64)


def b64_encode_local_file(path: str):
    mediatype = mimetypes.guess_type(path)[0]
    with open(path, 'rb') as fin:
        return b64_encode_data_url(mediatype, fin.read())


class Archive:
    @cached_property
    def inner_paths(self) -> List[str]:
        raise NotImplementedError

    @cached_property
    def entry_path(self) -> str:
        paths = [p for p in self.inner_paths if p.endswith('index.html')]
        if not paths:
            return ''
        return min(paths, key=lambda s: len(s))

    def extract_as_bytes(self, inner_path: str) -> bytes:
        raise NotImplementedError

    def open_inner_file(self, inner_path: str):
        raise NotImplementedError

    @staticmethod
    def open(path: str) -> 'Archive':
        if path.endswith('.zip'):
            return ZipArchive(path)
        tar_suffixes = ['.tar', '.tgz', '.tar.gz']
        for suffix in tar_suffixes:
            if path.endswith(suffix):
                return TarArchive(path)

    def __enter__(self):
        archive_file = getattr(self, 'archive_file', None)
        if hasattr(archive_file, '__enter__'):
            archive_file.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        archive_file = getattr(self, 'archive_file', None)
        if hasattr(archive_file, '__exit__'):
            archive_file.__exit__(exc_type, exc_val, exc_tb)
        return self


class TarArchive(Archive):
    def __init__(self, path: str):
        self.archive_file = tarfile.open(path)

    @cached_property
    def inner_paths(self) -> List[str]:
        return self.archive_file.getnames()

    def extract_as_bytes(self, inner_path: str) -> bytes:
        return self.archive_file.extractfile(inner_path).read()

    def open_inner_file(self, inner_path: str):
        return self.archive_file.extractfile(inner_path)


class ZipArchive(Archive):
    def __init__(self, path: str):
        self.archive_file = zipfile.ZipFile(path)

    def __enter__(self):
        self.archive_file.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.archive_file.__exit__(exc_type, exc_val, exc_tb)

    @cached_property
    def inner_paths(self) -> List[str]:
        return self.archive_file.namelist()

    def extract_as_bytes(self, inner_path: str) -> bytes:
        return self.archive_file.open(inner_path).read()

    def open_inner_file(self, inner_path: str):
        return self.archive_file.open(inner_path)


class Directory:
    def __init__(self, base_dir: str):
        self.base_dir = os.path.abspath(base_dir)

    def __repr__(self):
        c = self.__class__.__name__
        return '{}({!r})'.format(c, self.base_dir)

    def under(self, *paths):
        return os.path.join(self.base_dir, *paths)

    def relative_to_base_dir(self, path: str):
        path = os.path.abspath(path)
        return os.path.relpath(path, self.base_dir)

    under_base_dir = under

    def read_as_chunks(self, path: str, length=-1, offset=0, chunksize=65536) \
            -> Generator[bytes, None, None]:
        path = self.under_base_dir(path)
        return read_as_chunks(
            path, length=length,
            offset=offset, chunksize=chunksize,
        )

    def checksum_hexdigest(self, path: str, algo='sha1') -> str:
        path = self.under_base_dir(path)
        return checksum_hexdigest(path, algo=algo)

    def read_as_binary(self, path: str):
        path = self.under_base_dir(path)
        with open(path, 'rb') as fin:
            return fin.read()

    def read_as_base64_data_url(self, path: str):
        path = self.under_base_dir(path)
        return b64_encode_local_file(path)

    def save_as_file(self, path: str, chunks):
        path = self.under_base_dir(path)
        with open(path, 'wb') as fout:
            for chunk in chunks:
                fout.write(chunk)


class MappedDirectory(Directory):
    def __init__(self, base_dir: str, base_url: str):
        super().__init__(base_dir)
        # Note:
        # urllib.parse.urljoin('/a/b', 'c.jpg') => '/a/c.jpg'
        # urllib.parse.urljoin('/a/b/', 'c.jpg') => '/a/b/c.jpg'
        if not base_url.endswith('/'):
            base_url += '/'
        self.base_url = base_url

    def __repr__(self):
        c = self.__class__.__name__
        return '{}({!r}, {!r})'.format(c, self.base_dir, self.base_url)

    def join_url(self, path: str):
        return urllib.parse.urljoin(self.base_url, path)

    def relative_to_base_url(self, url: str):
        base_url_path = urllib.parse.urlparse(self.base_url).path
        url_path = urllib.parse.urlparse(url).path
        return os.path.relpath(url_path, base_url_path)

    def convert_local_path_to_url(self, path: str):
        path = os.path.abspath(path)
        path = self.relative_to_base_dir(path)
        return self.join_url(path)

    def convert_url_to_local_path(self, url: str):
        path = self.relative_to_base_url(url)
        return self.under(path)



#!/usr/bin/env python3
# coding: utf-8
from __future__ import annotations

import json
import logging
import shlex
import traceback
import urllib.parse
from json import JSONDecodeError
from typing import Union
from urllib.parse import urljoin

import requests
from volkanic.errors import TechnicalError
from volkanic.introspect import razor

_logger = logging.getLogger(__name__)


class ResponseDict(dict):
    @property
    def code(self) -> int:
        return self.get('code', 3)

    @property
    def data(self):
        return self.get('data')

    @property
    def message(self):
        return self.get('message', 'OK')


def dump_json_request_to_curl(method: str, url: str, data=None, aslist=False):
    method = method.upper()
    if method == 'GET':
        parts = ['curl', url]
    else:
        parts = [
            'curl', '-X', method, url,
            '-H', 'Content-Type: application/json',
            '-d', json.dumps(razor(data), ensure_ascii=False),
        ]
    if aslist:
        return parts
    parts = [shlex.quote(s) for s in parts]
    return ' '.join(parts)


def _log_bad_response(resp: requests.Response):
    _logger.error(
        'bad response: %s %r',
        resp.status_code, resp.content[:1000]
    )


def _decode_response(resp: requests.Response):
    status = resp.status_code
    if status >= 400:
        raise TechnicalError(f'got response status code {status}')
    try:
        rd = ResponseDict(resp.json())
    except JSONDecodeError:
        raise TechnicalError('cannot decode json')
    if rd.code != 0:
        raise TechnicalError(f'error response ({rd.code})')
    return rd.data


def decode_response(resp: requests.Response):
    try:
        return _decode_response(resp)
    except TechnicalError:
        _log_bad_response(resp)
        raise


class HTTPClient:
    timeout = 30

    def __init__(self, base_url: str):
        self.base_url = base_url

    def get_url(self, path: str):
        return urllib.parse.urljoin(self.base_url, path)

    @staticmethod
    def _load_json(resp: requests.Response) -> Union[dict, list, None]:
        err = _logger.error
        if resp.status_code != 200:
            err('status_code %s from %s', resp.status_code, resp.url)
            return
        try:
            data = json.loads(resp.text)
        except json.JSONDecodeError:
            traceback.print_exc()
            err('bad json from %s', resp.url)
            return
        if not isinstance(data, dict):
            err('wrong response type (%s) from %s', type(data), resp.url)
        code = data.get('code')
        if data.get('code'):
            err('non-zero code (%s) from %s', code, resp.url)
            return
        return data

    @staticmethod
    def log_request_as_curl(
            method: str, url: str, data=None, level=logging.DEBUG):
        if not _logger.isEnabledFor(level):
            return
        cmd = dump_json_request_to_curl(method, url, data=data)
        _logger.log(level, cmd)

    @staticmethod
    def log_resp_content(
            resp: requests.Response, level=logging.ERROR):
        if not _logger.isEnabledFor(level):
            return
        try:
            _logger.log(level, resp.text[:255])
        except Exception:
            _logger.log(level, resp.content[:255])
            traceback.print_exc()

    def json_request(self, method, path, data=None, **kwargs) \
            -> Union[dict, list, None]:
        url = self.get_url(path)
        method = method.upper()
        kwargs.setdefault('timeout', self.timeout)
        if data is not None:
            kwargs['data'] = json.dumps(data, default=str)
            kwargs['headers'] = {'Content-type': 'application/json'}
        resp = requests.request(method, url, **kwargs)
        resp_data = self._load_json(resp)
        if resp_data is None:
            self.log_resp_content(resp)
            self.log_request_as_curl(method, url, data)
            return
        self.log_request_as_curl(method, url, data, level=logging.DEBUG)
        return resp_data

    def json_get(self, path, **kwargs) -> Union[dict, list, None]:
        assert 'data' not in kwargs
        return self.json_request('GET', path, **kwargs)

    def json_post(self, path, data=None, **kwargs) -> Union[dict, list, None]:
        assert 'data' not in kwargs
        return self.json_request('POST', path, data=data, **kwargs)


HTTPService = HTTPClient


class HTTPClientInterface:
    def __init__(self, config: dict):
        self._upstream_config = config

    def __getitem__(self, service_name: str) -> HTTPService:
        return HTTPService(self._upstream_config[service_name])

    def get_service_url(self, service_name: str, path: str = None):
        base_url = self._upstream_config[service_name]
        if not path:
            return base_url
        return urljoin(base_url, path)


HTTPServiceInterface = HTTPClientInterface
